import os
import unittest
from typing import List, Tuple

from parameterized import parameterized

from pdfconduit import Pdfconduit
from pdfconduit.convert import PDF2IMG
from pdfconduit.convert.pdf2img import ImageExtension
from tests import *


def params() -> List[Tuple[str, ImageExtension, bool]]:
    files = [
        "article.pdf",
        "charts.pdf",
        "document.pdf",
        "plan_p.pdf",
    ]
    extensions = [ImageExtension.PNG, ImageExtension.JPG]
    from_stream = [True, False]

    return [
        (test_data_path(file), extension, use_stream)
        for use_stream in from_stream
        for extension in extensions
        for file in files
    ]


def name_func(testcase_func, param_num, param):
    return "{}.{}.{}.{}".format(
        testcase_func.__name__,  # test func
        param.args[1].value,  # image extension
        "from_stream" if param.args[2] is True else "from_file",  # stream vs. file
        get_clean_pdf_name(param.args[0]),  # filename
    )


class TestPdf2Img(PdfconduitTestCase):
    @parameterized.expand(params, name_func=name_func)
    def test_convert_pdf_to_images_pdf2img(
        self, pdf: str, ext: ImageExtension, from_stream: bool
    ):
        pdf2img = PDF2IMG(
            self._get_pdf(pdf, from_stream), output_directory=self.temp.name, ext=ext
        )
        images = pdf2img.convert()

        for image in images:
            # Assert img file exists
            self.assertTrue(os.path.exists(image))

            # Assert img file is correct file type
            self.assertTrue(image.endswith("." + ext.value))

    @parameterized.expand(params, name_func=name_func)
    def test_convert_pdf_to_images_pdfconduit(
        self, pdf: str, ext: ImageExtension, from_stream: bool
    ):
        self.conduit = Pdfconduit(self._get_pdf(pdf, from_stream)).set_output_temp(
            self.temp
        )
        images = self.conduit.to_images(ext=ext)

        for image in images:
            # Assert img file exists
            self.assertTrue(os.path.exists(image))

            # Assert img file is correct file type
            self.assertTrue(image.endswith("." + ext.value))

    @parameterized.expand(params, name_func=name_func)
    def test_convert_pdf_to_images_pdfconduit_without_setting_directory(
        self, pdf: str, ext: ImageExtension, from_stream: bool
    ):
        self.conduit = Pdfconduit(self._get_pdf(pdf, from_stream))
        images = self.conduit.to_images(ext=ext)

        for image in images:
            # Assert img file exists
            self.assertTrue(os.path.exists(image))

            # Assert img file is correct file type
            self.assertTrue(image.endswith("." + ext.value))

    def test_cant_convert_pdf_to_images_without_dir(self):
        with self.assertRaises(TypeError) as context:
            pdf2img = PDF2IMG(
                self._get_pdf(self.pdf_path, False), ext=ImageExtension.PNG
            )

        self.assertTrue(
            "missing 1 required positional argument: 'output_directory'"
            in str(context.exception)
        )

    def _get_pdf(self, pdf: str, from_stream: bool):
        return pdf if not from_stream else self._get_pdf_byte_stream(pdf)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from tempfile import NamedTemporaryFile, TemporaryDirectory

from pdfconduit import Info
from pdfconduit.convert import IMG2PDF
from tests import *


class TestImg2Pdf(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.img_path = img_path
        cls.pdf = None

    def setUp(self):
        self.tempdir = TemporaryDirectory()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_convert(self):
        with NamedTemporaryFile(suffix=".pdf", dir=self.tempdir.name) as tempfile:
            ip = IMG2PDF(self.img_path, output=tempfile.name, tempdir=self.tempdir)
            self.pdf = ip.convert()

            # Assert pdf file exists
            self.assertTrue(os.path.exists(self.pdf))
            self.assertEqual(1, Info(self.pdf).pages)
        ip.cleanup()

    def test_convert_packet(self):
        with NamedTemporaryFile(suffix=".pdf", dir=self.tempdir.name) as tempfile:
            ip = IMG2PDF(
                [self.img_path, self.img_path, self.img_path],
                output=tempfile.name,
                tempdir=self.tempdir,
            )
            self.pdf = ip.convert()

            # Assert pdf file exists
            self.assertTrue(os.path.exists(self.pdf))
            self.assertEqual(3, Info(self.pdf).pages)
        ip.cleanup()


if __name__ == "__main__":
    unittest.main()



# Convert each page of PDF to images
import os
from enum import Enum
from io import BytesIO
from typing import List, Optional

import fitz
from PIL import Image

try:
    from pymupdf import DisplayList
    from pymupdf import Document as PyMupdfDocument
except ImportError:
    from fitz import Document as PyMupdfDocument, DisplayList

from pdfconduit.utils.path import add_suffix
from pdfconduit.utils.typing import PdfObject


class ImageExtension(Enum):
    PNG: str = "png"
    JPG: str = "jpg"


class PDF2IMG:
    _filename: str = None
    _doc: PyMupdfDocument

    def __init__(
        self,
        pdf: PdfObject,
        output_directory: str,
        ext: ImageExtension = ImageExtension.PNG,
        alpha: bool = False,
    ):
        self._pdf = pdf
        self._directory = output_directory
        self._ext = ext.value
        self._alpha = alpha

        if isinstance(self._pdf, BytesIO):
            self._filename = "pdf2img"
            self._doc = fitz.open(stream=self._pdf)
        else:
            self._filename = os.path.basename(self._pdf)
            self._doc = fitz.open(filename=self._pdf)

    def convert(self) -> list:
        saved = []
        for index, image in enumerate(self._get_pdf_data()):
            output = os.path.join(
                self._directory,
                add_suffix(self._filename, str(index + 1), ext=self._ext),
            )
            with Image.open(BytesIO(image)) as pillow:
                pillow.save(output)
            saved.append(output)
        self._doc.close()
        return saved

    def _get_pdf_data(self) -> List[bytes]:
        return FitzPdfToImage(self._doc, self._alpha).convert()


class FitzPdfToImage:
    def __init__(self, doc: PyMupdfDocument, alpha: bool = False):
        self._doc = doc
        self._dlist_tab = [None] * len(doc)
        self._alpha = alpha

    def convert(self) -> List[bytes]:
        return [self._get_page_data(cur_page) for cur_page in range(len(self._doc))]

    def _get_page_data(self, pno, zoom=0) -> bytes:
        """
        Return a PNG image for a document page number. If zoom is other than 0, one of
        the 4 page quadrants are zoomed-in instead and the corresponding clip returned.
        """
        dlist: Optional[DisplayList] = self._dlist_tab[pno]  # get display list
        if not dlist:  # create if not yet there
            self._dlist_tab[pno] = self._doc[pno].get_displaylist()
            dlist = self._dlist_tab[pno]
        r = dlist.rect  # page rectangle
        mp = r.tl + (r.br - r.tl) * 0.5  # rect middle point
        mt = r.tl + (r.tr - r.tl) * 0.5  # middle of top edge
        ml = r.tl + (r.bl - r.tl) * 0.5  # middle of left edge
        mr = r.tr + (r.br - r.tr) * 0.5  # middle of right egde
        mb = r.bl + (r.br - r.bl) * 0.5  # middle of bottom edge
        mat = fitz.Matrix(2, 2)  # zoom matrix
        if zoom == 1:  # top-left quadrant
            clip = fitz.Rect(r.tl, mp)
        elif zoom == 4:  # bot-right quadrant
            clip = fitz.Rect(mp, r.br)
        elif zoom == 2:  # top-right
            clip = fitz.Rect(mt, mr)
        elif zoom == 3:  # bot-left
            clip = fitz.Rect(ml, mb)
        if zoom == 0:  # total page
            pix = dlist.get_pixmap(alpha=self._alpha)
        else:
            pix = dlist.get_pixmap(alpha=self._alpha, matrix=mat, clip=clip)
        return pix.tobytes()  # return the PNG image



# Convert a PNG image file to a PDF
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional, Union

from PIL import Image

from pdfconduit.transform import Merge2
from pdfconduit.watermark.modify.canvas import CanvasImg, CanvasObjects
from pdfconduit.watermark.modify.draw import WatermarkDraw


class IMG2PDF:
    def __init__(
        self,
        images: Union[str, List[str]],
        output: Optional[str] = None,
        tempdir: Optional[TemporaryDirectory] = None,
    ):
        """Convert each image into a PDF page and merge all pages to one PDF file"""
        self.images = images if isinstance(images, list) else [images]
        self.output = output
        self._tempdir = tempdir or TemporaryDirectory()

    def convert(self, cleanup_temp: bool = True) -> str:
        pdfs = [self._convert_image_to_pdf(image) for image in self.images]
        merged = Merge2(pdfs, output=self.output).merge()
        return merged

    def cleanup(self) -> None:
        self._tempdir.cleanup()

    def _convert_image_to_pdf(self, image: str) -> str:
        with Image.open(image) as im:
            width, height = im.size

            co = CanvasObjects()
            co.add(CanvasImg(image, 1.0, w=width, h=height, mask=None))

            return WatermarkDraw(
                co, tempdir=self._tempdir.name, pagesize=(width, height)
            ).write(
                os.path.join(
                    self._tempdir.name,
                    os.path.basename(image.replace(Path(image).suffix, ".pdf")),
                )
            )



# Rotate a pdf file
import os
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import Optional

from pdfrw import PdfReader as PdfrwReader
from pdfrw import PdfWriter as PdfrwWriter
from pypdf import PdfReader as PypdfReader
from pypdf import PdfWriter as PypdfWriter

from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.path import add_suffix
from pdfconduit.utils.typing import PdfObject


class Rotate(PdfDriver):

    def __init__(
        self,
        pdf: PdfObject,
        rotation: int,
        suffix: str = "rotated",
        tempdir: Optional[str] = None,
        output: Optional[str] = None,
    ):
        self.pdf_object = pdf
        self.rotation = rotation
        self.suffix = suffix

        if output:
            self.outfn = output
        else:
            self.tempdir = tempdir

            if tempdir:
                with NamedTemporaryFile(
                    suffix=".pdf", dir=tempdir, delete=False
                ) as temp:
                    self.outfn = temp.name
            elif suffix:
                self.outfn = os.path.join(os.path.dirname(pdf), add_suffix(pdf, suffix))
            else:
                self.outfn = NamedTemporaryFile(suffix=".pdf").name

    def __str__(self) -> str:
        return self.file

    def rotate(self) -> str:
        return self.execute()

    @property
    def file(self) -> str:
        return str(self.outfn)

    def pdfrw(self) -> str:
        if isinstance(self.pdf_object, BytesIO):
            trailer = PdfrwReader(fdata=self.pdf_object.getvalue())
        else:
            trailer = PdfrwReader(fname=self.pdf_object)

        pages = trailer.pages

        ranges = [[1, len(pages)]]

        for onerange in ranges:
            onerange = (onerange + onerange[-1:])[:2]
            for pagenum in range(onerange[0] - 1, onerange[1]):
                pages[pagenum].Rotate = (
                    int(pages[pagenum].inheritable.Rotate or 0) + self.rotation
                ) % 360

        outdata = PdfrwWriter(self.outfn)
        outdata.trailer = trailer
        outdata.write()
        return self.outfn

    def pypdf(self) -> str:
        reader = PypdfReader(self.pdf_object)
        writer = PypdfWriter()

        for page_num in range(1, reader.get_num_pages()):
            writer.add_page(reader.pages[page_num]).rotate(self.rotation)

        with open(self.outfn, "wb") as fp:
            writer.write(fp)

        return self.outfn



# Merge PDF documents
import os
from typing import List, Optional, Union

from pdfrw import IndirectPdfDict as PdfrwIndirectPdfDict
from pdfrw import PdfReader as PdfrwReader
from pdfrw import PdfWriter as PdfrwWriter
from pypdf import PdfWriter as PyPdfWriter

from pdfconduit.utils.driver import PdfDriver

INPUT_PDFS_TYPE = Union[list, str]


class Merge(PdfDriver):
    file: str = None

    def __init__(
        self,
        input_pdfs: INPUT_PDFS_TYPE,
        output_name: str = "merged",
        output_dir: Optional[str] = None,
    ):
        self.pdfs = self._get_pdf_list(input_pdfs)
        self.directory = output_dir if output_dir else os.path.dirname(self.pdfs[0])
        self.output = os.path.join(
            self.directory, output_name.replace(".pdf", "") + ".pdf"
        )

    def __str__(self) -> str:
        return str(self.file)

    @staticmethod
    def validate(pdf) -> bool:
        # todo: simplify logic?
        if not pdf.startswith(".") and pdf.endswith(".pdf"):
            return True

    def _get_pdf_list(self, input_pdfs: INPUT_PDFS_TYPE) -> List[str]:
        """
        Generate list of PDF documents.

        :param input_pdfs: List of PDFs or a directory path
             Directory - Scans directory contents
             List - Filters list to assert all list items are paths to PDF documents
        :return: List of PDF paths
        """
        if isinstance(input_pdfs, list):
            return [pdf for pdf in input_pdfs if self.validate(pdf)]
        elif os.path.isdir(input_pdfs):
            return [
                os.path.join(input_pdfs, pdf)
                for pdf in os.listdir(input_pdfs)
                if self.validate(pdf)
            ]
        # todo: raise error if conditions aren't met?

    def merge(self) -> str:
        """Merge list of PDF files to a single PDF file."""
        self.file = self.execute()
        return self.file

    def pdfrw(self):
        writer = PdfrwWriter()
        for inpfn in self.pdfs:
            writer.addpages(PdfrwReader(inpfn).pages)

        writer.trailer.Info = PdfrwIndirectPdfDict(
            Author="Stephen Neal",
            Creator="pdfconduit",
            Producer="pdfconduit",
        )
        writer.write(self.output)
        return self.output

    def pypdf(self):
        merger = PyPdfWriter()

        for pdf in self.pdfs:
            merger.append(pdf)

        merger.write(self.output)
        merger.close()

        return self.output



# Upscale a PDF file
from io import BytesIO

from pdfrw import IndirectPdfDict as pdfrwIndirectPdfDict
from pdfrw import PageMerge as pdfrwPageMerge
from pdfrw import PdfReader as pdfrwReader
from pdfrw import PdfWriter as pdfrwWriter
from pypdf import PdfReader as pypdfReader
from pypdf import PdfWriter as pypdfWriter

from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.info import Info
from pdfconduit.utils.typing import PdfObject


class Scale(PdfDriver):
    def __init__(
        self,
        pdf: PdfObject,
        output: str,
        scale: float = 1.5,
        margin_x: int = 0,
        margin_y: int = 0,
    ):
        self._pdf = pdf
        self._output = output
        self._margin_x = margin_x
        self._margin_y = margin_y
        self._scale = scale

    def upscale(self) -> str:
        # Execute either pdfrw or PyPDF3 method
        return self.execute()

    def _pdfrw_adjust(self, page: object):
        info = pdfrwPageMerge().add(page)
        x1, y1, x2, y2 = info.xobj_box
        viewrect = (
            self._margin_x,
            self._margin_y,
            x2 - x1 - 2 * self._margin_x,
            y2 - y1 - 2 * self._margin_y,
        )
        page = pdfrwPageMerge().add(page, viewrect=viewrect)
        page[0].scale(self._scale)
        return page.render()

    def pdfrw(self) -> str:
        if isinstance(self._pdf, BytesIO):
            reader = pdfrwReader(fdata=self._pdf.getvalue())
        else:
            reader = pdfrwReader(fname=self._pdf)

        writer = pdfrwWriter(self._output)

        for i in list(range(0, len(reader.pages))):
            writer.addpage(self._pdfrw_adjust(reader.pages[i]))

        writer.trailer.Info = pdfrwIndirectPdfDict(reader.Info or {})
        writer.write()

        return self._output

    def pypdf(self) -> str:
        reader = pypdfReader(self._pdf)

        # Get target width and height
        dims = Info(reader).dimensions
        target_w = dims["w"] * self._scale
        target_h = dims["h"] * self._scale

        writer = pypdfWriter()

        for page_num in range(0, reader.get_num_pages()):
            page = reader.pages[page_num]
            page.scale_to(width=target_w, height=target_h)
            writer.add_page(page)

        with open(self._output, "wb") as fp:
            writer.write(fp)

        return self._output



# Merge PDF documents
from io import BytesIO

from pdfrw import IndirectPdfDict as PdfrwIndirectPdfDict
from pdfrw import PdfReader as PdfrwReader
from pdfrw import PdfWriter as PdfrwWriter
from pypdf import PdfWriter as PyPdfWriter

from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.typing import PdfObjects


class Merge2(PdfDriver):
    def __init__(self, pdfs: PdfObjects, output: str):
        self._pdfs = pdfs
        self._output = output

    def merge(self) -> str:
        """Merge list of PDF files to a single PDF file."""
        return self.execute()

    def pdfrw(self):
        writer = PdfrwWriter()

        for pdf_object in self._pdfs:
            if isinstance(pdf_object, BytesIO):
                reader = PdfrwReader(fdata=pdf_object.getvalue())
            else:
                reader = PdfrwReader(fname=pdf_object)
            writer.addpages(reader.pages)

        writer.trailer.Info = PdfrwIndirectPdfDict(
            Author="Stephen Neal",
            Creator="pdfconduit",
            Producer="pdfconduit",
        )
        writer.write(self._output)
        return self._output

    def pypdf(self):
        merger = PyPdfWriter()

        for pdf in self._pdfs:
            merger.append(pdf)

        merger.write(self._output)
        merger.close()

        return self._output



# Upscale a PDF file
import os
from tempfile import NamedTemporaryFile
from typing import Optional

from pdfrw import IndirectPdfDict as pdfrwIndirectPdfDict
from pdfrw import PageMerge as pdfrwPageMerge
from pdfrw import PdfReader as pdfrwReader
from pdfrw import PdfWriter as pdfrwWriter
from pypdf import PdfReader as pypdfReader
from pypdf import PdfWriter as pypdfWriter

from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.info import Info
from pdfconduit.utils.path import add_suffix


class Upscale(PdfDriver):
    def __init__(
        self,
        file_name: str,
        margin_x: int = 0,
        margin_y: int = 0,
        scale: float = 1.5,
        suffix: str = "scaled",
        tempdir: Optional[str] = None,
    ):
        self.file_name = file_name
        self.margin_x = margin_x
        self.margin_y = margin_y
        self.scale = scale
        self.suffix = suffix

        # Set output file name
        if tempdir:
            with NamedTemporaryFile(
                suffix="_" + suffix + ".pdf", dir=tempdir, delete=False
            ) as temp:
                self.output = temp.name
        elif suffix:
            self.output = os.path.join(
                os.path.dirname(file_name), add_suffix(file_name, suffix)
            )
        else:
            self.output = NamedTemporaryFile(suffix="_" + suffix + ".pdf").name

        # Get target width and height
        dims = Info(self.file_name).dimensions
        self.target_w = dims["w"] * self.scale
        self.target_h = dims["h"] * self.scale

    def __str__(self) -> str:
        return self.file

    def upscale(self) -> str:
        # Execute either pdfrw or PyPDF3 method
        return self.execute()

    @property
    def file(self) -> str:
        return str(self.output)

    def _pdfrw_adjust(self, page: object):
        info = pdfrwPageMerge().add(page)
        x1, y1, x2, y2 = info.xobj_box
        viewrect = (
            self.margin_x,
            self.margin_y,
            x2 - x1 - 2 * self.margin_x,
            y2 - y1 - 2 * self.margin_y,
        )
        page = pdfrwPageMerge().add(page, viewrect=viewrect)
        page[0].scale(self.scale)
        return page.render()

    def pdfrw(self) -> str:
        reader = pdfrwReader(self.file_name)
        writer = pdfrwWriter(self.output)
        for i in list(range(0, len(reader.pages))):
            writer.addpage(self._pdfrw_adjust(reader.pages[i]))
        writer.trailer.Info = pdfrwIndirectPdfDict(reader.Info or {})
        writer.write()

        return self.output

    def pypdf(self) -> str:
        reader = pypdfReader(self.file_name)
        writer = pypdfWriter()

        for page_num in range(0, reader.get_num_pages()):
            page = reader.pages[page_num]

            page.scale_to(width=self.target_w, height=self.target_h)

            writer.add_page(page)

        with open(self.output, "wb") as fp:
            writer.write(fp)

        return self.output



# Apply a watermark to a PDF file
import os
import shutil
from tempfile import TemporaryDirectory
from typing import Optional

from looptools import Timer

from pdfconduit.main import Pdfconduit
from pdfconduit.settings import Encryption
from pdfconduit.utils import Info, add_suffix
from pdfconduit.watermark.add import WatermarkAdd
from pdfconduit.watermark.lib import IMAGE_DEFAULT, IMAGE_DIRECTORY
from pdfconduit.watermark.modify.canvas import CanvasConstructor
from pdfconduit.watermark.modify.draw import WatermarkDraw


class Watermark:
    def __init__(
        self,
        document: str,
        remove_temps: bool = True,
        move_temps: Optional[bool] = None,
        tempdir: Optional[str] = None,
        use_receipt: bool = True,
    ):
        """
        Watermark and encrypt a PDF document.

        Manage watermarking processes from single class initialization.  This class utilizes the draw,
        add and encrypt modules.

        :param document: str
            PDF document full path
        :param remove_temps: bool
            Remove temporary files after completion
        :param open_file: bool
            Open file after completion
        :param tempdir: function or str
            Temporary directory for file writing
        :param receipt: cls
            Use existing Receipt object if already initiated
        :param use_receipt: bool
            Print receipt information to console and write to file
        """
        self.time = Timer()
        self.document_og = document
        self.document = self.document_og
        self.watermark = None
        self.remove_temps = remove_temps
        self.move_temps = move_temps

        if not tempdir:
            self._temp = TemporaryDirectory()
            self.tempdir = self._temp.name
        elif isinstance(tempdir, TemporaryDirectory):
            self._temp = tempdir
            self.tempdir = self._temp.name
        else:
            self.tempdir = tempdir

    def __str__(self) -> str:
        return str(self.document)

    def cleanup(self) -> str:
        self.time.end
        if self.move_temps:
            if os.path.isdir(self.move_temps):
                shutil.move(self.tempdir, self.move_temps)
        if self.remove_temps:
            if os.path.isdir(self.tempdir):
                shutil.rmtree(self.tempdir)
        return self.document

    def draw(
        self,
        text1: Optional[str] = None,
        text2: Optional[str] = None,
        include_copyright: bool = True,
        image: str = IMAGE_DEFAULT,
        rotate: int = 30,
        opacity: float = 0.08,
        compress: int = 0,
        flatten: bool = False,
        add: bool = False,
    ) -> str:
        """
        Draw watermark PDF file.

        Create watermark using either a reportlabs canvas or a PIL image.

        :param text1: str
            Text line 1
        :param text2: str
            Text line 2
        :param include_copyright: bool
            Draw copyright and year to canvas
        :param image: str
            Logo image to be used as base watermark
        :param rotate: int
            Degrees to rotate canvas by
        :param opacity: float
            Watermark opacity
        :param compress: bool
            Compress watermark contents  (not entire PDF)
        :param flatten: bool
            Draw watermark with multiple layers or a single flattened layer
        :param add: bool
            Add watermark to original document
        :return: str
            Watermark PDF file full path
        """
        im_path = os.path.join(IMAGE_DIRECTORY, image)
        if os.path.isfile(im_path):
            image = im_path

        co = CanvasConstructor(
            text1,
            text2,
            include_copyright,
            image,
            rotate,
            opacity,
            tempdir=self.tempdir,
        )
        objects, rotate = (
            co.img() if flatten else co.canvas()
        )  # Run img constructor method if flatten is True

        # Draw watermark to file
        self.watermark = WatermarkDraw(
            objects,
            rotate=rotate,
            compress=compress,
            tempdir=self.tempdir,
            pagesize=Info(self.document_og).size,
            pagescale=True,
        ).write()

        if not add:
            return self.watermark
        else:
            self.add()
            return self.cleanup()

    def add(
        self,
        document: Optional[str] = None,
        watermark: Optional[str] = None,
        underneath: bool = False,
        output: Optional[str] = None,
        suffix: Optional[str] = "watermarked",
        method: str = "pdfrw",
    ) -> str:
        """
        Add a watermark file to an existing PDF document.

        Rotate and upscale watermark file as needed to fit existing PDF document.  Watermark can be overlayed or
        placed underneath.

        :param document: str
            PDF document full path
        :param watermark: str
            Watermark PDF full path
        :param underneath: bool
            Place watermark either under or over existing PDF document
        :param output: str
            Output file path
        :param suffix: str
            Suffix to append to existing PDF document file name
        :param method: str
            PDF library to be used for watermark adding
        :return: str
            Watermarked PDF Document full path
        """
        if not watermark:
            watermark = self.watermark
        if not document:
            document = self.document

        watermarker = WatermarkAdd(
            document,
            watermark,
            output=output,
            underneath=underneath,
            tempdir=self.tempdir,
            suffix=suffix,
        )
        if method == "pdfrw":
            watermarker.use_pdfrw()
        else:
            watermarker.use_pypdf()
        self.document = watermarker.add()

        return self.document

    def encrypt(
        self,
        user_pw: str = "",
        owner_pw: Optional[str] = None,
        encrypt_128: bool = True,
        allow_printing: bool = True,
        allow_commenting: bool = False,
        document: Optional[str] = None,
    ):
        """
        Encrypt a PDF document to add passwords and restrict permissions.

        Add a user password that must be entered to view document and a owner password that must be entered to alter
        permissions and security settings.  Encryption keys are 128 bit when encrypt_128 is True and 40 bit when
        False.  By default permissions are restricted to print only, when set to false all permissions are allowed.
        TODO: Add additional permission parameters

            User password required to open and view PDF document
        :param owner_pw: str
            Owner password required to alter security settings and permissions
        :param encrypt_128: bool
            Encrypt PDF document using 128 bit keys
        :param allow_printing: bool
            Restrict permissions to print only
        :param allow_commenting:
        :param user_pw: str
        :param document:
        :return: str
            Encrypted PDF full path
        """
        document = self.document if document is None else document

        encrypter = Encryption(
            user_pw=user_pw,
            owner_pw=owner_pw,
            allow_printing=allow_printing,
            allow_commenting=allow_commenting,
        )
        p = (
            Pdfconduit(document)
            .set_output(add_suffix(self.document_og, "secured"))
            .encrypt(encrypter)
            .write()
        )

        return p



from typing import List

from parameterized import parameterized

from pdfconduit import Info, Pdfconduit
from tests import PdfconduitTestCase, get_clean_pdf_name, test_data_path


def decrypt_params() -> List[str]:
    return list(
        map(
            lambda filename: test_data_path(filename),
            [
                "article_encrypted.pdf",
                "manual_encrypted.pdf",
                "workbook_encrypted.pdf",
            ],
        )
    )


def decrypt_name_func(testcase_func, param_num, param):
    return "{}.{}".format(
        testcase_func.__name__,
        get_clean_pdf_name(param.args[0]),
    )


class TestDecrypt(PdfconduitTestCase):
    @parameterized.expand(decrypt_params, name_func=decrypt_name_func)
    def test_can_decrypt(self, path: str):
        self.assertTrue(Info(path).encrypted)
        self.conduit = (
            Pdfconduit(path, "foo")
            .set_output_directory(self.temp.name)
            .set_output_suffix("decrypted")
        )
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertFalse(self.conduit.info.encrypted)
        self.assertTrue(self.conduit.info.decrypted)

        # Double check
        self.assertFalse(Info(self.conduit.output).encrypted)
        self.assertTrue(Info(self.conduit.output).decrypted)
